#!/usr/bin/env python
# -*- coding: utf-8 -*-

## MoonShineLua (MSL)
##
## Copyright 2016 Andrei Costin (andrei@firmware.re) for Firmware.RE
## with support from Tjaldur Software Governance Solutions and NLnet.nl.
##
## See LICENSE file for details.

import re

import msl.config.common as commonconfig

# TODO: for "config_db"
# have these more structured definitions (perhaps as Class) 
# or add more annotations, containing meta-data such as:
#   vuln category (e.g., "XSS")
#   CWE id and description
#   default sanitization suggestion (e.g., htmlescape)
#   default exploit generator function/class, perhaps taking as inputs
#       contraints (e.g., which chars are sanitized)
#       context (e.g., concatenated/placed with/within single-quotes/double-quotes strings)
#       contraint solver (e.g., SAT/SMT)

# TODO: DOC: better HOWTO: add a "config_db" entry

# TODO: DOC: theory/idea behind each field (e.g., "passthru")

config_db = [
    {
        "id": "_GET", 
        "match": re.compile("^_GET\["), 
        "desc": "", 
        "ref": "http://www.ntop.org/ntopng/scripting-ntopng-with-lua/", 
        "notes": "", 
        "is_tainted": True, 
        "taint_types": commonconfig.DEFAULT_VULN_SET, 
        "passthru": commonconfig.DEFAULT_VULN_SET, 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set([]), 
    },
    {
        "id": "cgilua.QUERY", 
        "match": re.compile("^cgilua.QUERY."), 
        "desc": "", 
        "ref": "https://keplerproject.github.io/cgilua/manual.html", 
        "notes": "", 
        "is_tainted": True, 
        "taint_types": commonconfig.DEFAULT_VULN_SET, 
        "passthru": commonconfig.DEFAULT_VULN_SET, 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set([]), 
    },
    {
        "id": "cgilua.POST", 
        "match": re.compile("^cgilua.POST."), 
        "desc": "", 
        "ref": "https://keplerproject.github.io/cgilua/manual.html", 
        "notes": "", 
        "is_tainted": True, 
        "taint_types": commonconfig.DEFAULT_VULN_SET, 
        "passthru": commonconfig.DEFAULT_VULN_SET, 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set([]), 
    },
    {
        "id": "print", 
        "match": re.compile("^print$"), 
        "desc": "Receives any number of arguments and prints their values to stdout, using the tostring function to convert each argument to a string.", 
        "ref": "https://www.lua.org/manual/5.3/manual.html#pdf-print", 
        "notes": "", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": set([]), 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set(["xss"]), 
    },
    {
        "id": "io.write", 
        "match": re.compile("^io.write$"), 
        "desc": "Writes the value of each of its arguments to file. In case of success, this function returns file. Otherwise it returns nil plus a string describing the error.", 
        "ref": "https://www.lua.org/manual/5.3/manual.html#pdf-io.write", 
        "notes": "", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": set([]), 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set(["xss"]), 
    },
    {
        "id": "io.open", 
        "match": re.compile("^io.open$"), 
        "desc": "Opens a file, in the mode specified in the string mode. In case of success, it returns a new file handle.", 
        "ref": "https://www.lua.org/manual/5.3/manual.html#pdf-io.open", 
        "notes": "", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": set([]), 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set(["pathtrav", "nullbyte", ]), 
    },
    {
        "id": "cgilua.put", 
        "match": re.compile("^cgilua.put$"), 
        "desc": "Function that outputs to stdout the string arguments it receives. Returns nothing.", 
        "ref": "https://keplerproject.github.io/cgilua/reference.html", 
        "notes": "", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": set([]), 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set(["xss"]), 
    },
    {
        "id": "cgilua.handlelp", 
        "match": re.compile("^cgilua.handlelp$"), 
        "desc": "Equivalent to cgilua.lp.include but sends the HTML header before the pre-processed file. Returns nothing.", 
        "ref": "https://keplerproject.github.io/cgilua/reference.html", 
        "notes": "", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": set([]), 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set(["lfi", "pathtrav", "nullbyte", ]), 
    },
    {
        "id": "execute", 
        "match": re.compile("^execute$"), 
        "desc": "", 
        "ref": "", 
        "notes": "", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": set([]), 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set(["sqli"]), 
    },
    {
        "id": "strcat", 
        "match": re.compile("^strcat$"), 
        "desc": "Function that concatenates two string arguments it receives. Returns the result of concatenation as string.", 
        "ref": "http://www.cplusplus.com/reference/cstring/strcat/", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET, 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set([]), 
    },
    {
        "id": "htmlentities", 
        "match": re.compile("^htmlentities$"), 
        "desc": "Convert all applicable characters to HTML entities. Returns the encoded string.", 
        "ref": "http://php.net/manual/en/function.htmlentities.php", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET.difference(set(["xss"])), 
        "unsanitize": set([]), 
        "sanitize": set(["xss"]), 
        "sensitive": set([]), 
    },
    {
        "id": "htmlspecialchars", 
        "match": re.compile("^htmlspecialchars$"), 
        "desc": "Convert special characters to HTML entities. Returns the converted string.", 
        "ref": "http://php.net/manual/en/function.htmlspecialchars.php", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET.difference(set(["xss"])), 
        "unsanitize": set([]), 
        "sanitize": set(["xss"]), 
        "sensitive": set([]), 
    },
    {
        "id": "htmlescape", 
        "match": re.compile("^htmlescape$"), 
        "desc": "Synonym for htmlentities.", 
        "ref": "", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET.difference(set(["xss"])), 
        "unsanitize": set([]), 
        "sanitize": set(["xss"]), 
        "sensitive": set([]), 
    },
    {
        "id": "html_entity_decode", 
        "match": re.compile("^html_entity_decode$"), 
        "desc": "Convert all HTML entities to their applicable characters. Returns the decoded string.", 
        "ref": "http://php.net/manual/en/function.html-entity-decode.php", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET, 
        "unsanitize": set(["xss"]), 
        "sanitize": set([]), 
        "sensitive": set([]), 
    },
    {
        "id": "htmlunescape", 
        "match": re.compile("^htmlunescape$"), 
        "desc": "Synonym for html_entity_decode.", 
        "ref": "", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET, 
        "unsanitize": set(["xss"]), 
        "sanitize": set([]), 
        "sensitive": set([]), 
    },
    {
        "id": "io.popen", 
        "match": re.compile("^io.popen$"), 
        "desc": "Starts program prog in a separated process. Returns a file handle that you can use to read data from this program.", 
        "ref": "https://www.lua.org/manual/5.1/manual.html#pdf-io.popen", 
        "notes": "", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": set([]), 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set(["cmdi"]), 
    },
    {
        "id": "escapeshellarg", 
        "match": re.compile("^escapeshellarg$"), 
        "desc": "Escape a string to be used as a shell argument. Returns the escaped string.", 
        "ref": "http://php.net/manual/en/function.escapeshellarg.php", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET.difference(set(["cmdi"])), 
        "unsanitize": set([]), 
        "sanitize": set(["cmdi"]), 
        "sensitive": set([]), 
    },
    {
        "id": "unescapeshellarg", 
        "match": re.compile("^unescapeshellarg$"), 
        "desc": "UNEscape a string to be used as a shell argument. Returns the UNescaped string.", 
        "ref": "FAKE", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET, 
        "unsanitize": set(["cmdi"]), 
        "sanitize": set([]), 
        "sensitive": set([]), 
    },
    {
        "id": "escapeshellcmd", 
        "match": re.compile("^escapeshellcmd$"), 
        "desc": "Escape shell metacharacters. Returns the escaped string.", 
        "ref": "http://php.net/manual/en/function.escapeshellcmd.php", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET.difference(set(["cmdi"])), 
        "unsanitize": set([]), 
        "sanitize": set(["cmdi"]), 
        "sensitive": set([]), 
    },
    {
        "id": "unescapeshellcmd", 
        "match": re.compile("^unescapeshellcmd$"), 
        "desc": "UNEscape shell metacharacters. Returns the UNescaped string.", 
        "ref": "FAKE", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET, 
        "unsanitize": set(["cmdi"]), 
        "sanitize": set([]), 
        "sensitive": set([]), 
    },
    {
        "id": "fake_htmlunescape_print_popen", 
        "match": re.compile("^fake_htmlunescape_print_popen\("), 
        "desc": "Hypothetical function that performs similar to htmlunescape, but also calls print/cgilua.put and io.popen within itself as part of its functionality.", 
        "ref": "FAKE", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET, 
        "unsanitize": set(["xss"]), 
        "sanitize": set([]), 
        "sensitive": set(["xss", "cmdi"]), 
    },
    {
        "id": "fake_htmlescape_print_popen", 
        "match": re.compile("^fake_htmlescape_print_popen\("), 
        "desc": "Hypothetical function that performs similar to htmlescape, but also calls print/cgilua.put and io.popen within itself as part of its functionality.", 
        "ref": "FAKE", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET.difference(set(["xss"])), 
        "unsanitize": set([]), 
        "sanitize": set(["xss"]), 
        "sensitive": set(["xss", "cmdi"]), 
    },
    {
        "id": "fake_strcat_print_popen", 
        "match": re.compile("^fake_strcat_print_popen\("), 
        "desc": "Hypothetical function that performs similar to strcat, but also calls print/cgilua.put and io.popen within itself as part of its functionality.", 
        "ref": "FAKE", 
        "notes": "NOTE: does not exist in Lua per-se, added for testing MSL", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": commonconfig.DEFAULT_VULN_SET, 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set(["xss", "cmdi"]), 
    },
    {
        "id": "ntop.readdir", 
        "match": re.compile("^ntop.readdir\("), 
        "desc": "Lists directories from a given path using opendir(path) and readdir(path). Returns integer number as CONST_LUA_ERROR or CONST_LUA_OK.", 
        "ref": "https://sourceforge.net/projects/ntop/files/ntopng/old/ntopng-1.1_6932.tgz#ntopng-1.1_6932/Lua.cpp", 
        "notes": "ntopng-1.1_6932/scripts/lua/modules/graph_utils.lua \n Related: CVE-2014-4329", 
        "is_tainted": False, 
        "taint_types": set([]), 
        "passthru": set([]), 
        "unsanitize": set([]), 
        "sanitize": set([]), 
        "sensitive": set(["pathtrav", "fileenum", ]), 
    },
]

config_db_as_dict = dict([(d["id"],d) for d in config_db for key in d])

