#!/usr/bin/env python
# -*- coding: utf-8 -*-

## MoonShineLua (MSL)
##
## Copyright 2016 Andrei Costin (andrei@firmware.re) for Firmware.RE
## with support from Tjaldur Software Governance Solutions and NLnet.nl.
##
## See LICENSE file for details.


'''
Unit tests for MSL.
'''

import os
import sys
import pprint
import unittest
import msl.tests.common as commontests
import msl.utils.lua_parser_json as mslparser
import msl.utils.utils as mslutils
import msl.taint.utils as taintutils

class CVE_2014_4329_Tests(unittest.TestCase):
    ############################################################################
    def setUp(self):
        (self.dir, self.dir_realpath) = commontests.get_tests_dir_prefix()
        self.code = ""
        self.clsname = self.__class__.__name__
        self.funcname = sys._getframe().f_code.co_name

    ############################################################################
    def tearDown(self):
        self.dir = None
        self.dir_realpath = None
        self.code = ""
        self.clsname = None
        self.funcname = None

    ############################################################################

    def test_xss(self):
        self.funcname = sys._getframe().f_code.co_name

        assert(self.funcname != None)
        assert(self.code != None)

        # NOTE: This is an external file with code for analysis
        host_details_lua = "./tests/3rdparty/ntopng-1.1_6932/scripts/lua/host_details.lua"
        test_fname = os.path.normpath(os.path.join(self.dir_realpath, host_details_lua))
        assert(test_fname != None)
        assert(os.path.exists(test_fname))
        assert(os.path.isfile(test_fname))

        luajsonres = mslparser.lua_get_json(test_fname)
        assert(luajsonres != None)

        cst_tlib = luajsonres["cst_tlib"]
        assert(cst_tlib != None)
        
        variables_all = luajsonres["variables_def"] + luajsonres["variables_local_def"]
        (variables_ordered, variables_independent, variables_dependent, variables_to_expressions_map) = taintutils.solve_variables_dependency_and_order(cst_tlib, variables_all)

        vulns_dict_all = {}

        ########################################################################
        
        for vname in variables_ordered:
            # Get the JSON object related to this variable
            v = [d for d in variables_all if d["text"] == vname][0]
            
            # Solve/taint current variable's (sub)tree
            (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, v["node"], v["text"], variables_to_expressions_map)
            
            # Merge current variable vulnerabilities dict into the all vulnerabilities dict
            vulns_dict_all.update(vulns_dict_var)
            
        ########################################################################
        
        cst_tlib_depth = cst_tlib.depth()
        all_nodes = cst_tlib.all_nodes()
        # We go just one level lower than the root node, therefore the second parameter is 0 instead of -1
        for depth in range(cst_tlib_depth, 0, -1):
            for node_tlib in all_nodes:
                node_solved = node_tlib.data.get("node_state", {"is_solved": False}).get("is_solved", False)
                if cst_tlib.level(node_tlib.identifier) == depth and not node_solved:
                    node_antlr = node_tlib.data["node_antlr"]
                    node_antlr_text = node_antlr.getText().encode('utf-8')
                    
                    # Solve/taint current unsolved node (sub)tree
                    (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, node_antlr, node_antlr_text, variables_to_expressions_map)
                    
                    # Merge current variable vulnerabilities dict into the all vulnerabilities dict
                    vulns_dict_all.update(vulns_dict_var)
                    
        ########################################################################

        vuln_types = set([])
        vuln_lines = set([])
        vuln_columns = set([])

        for vuln in vulns_dict_all.values():
            node_antlr = vuln[0]
            node_antlr_text = node_antlr.getText().encode('utf-8')
            type_of_vuln = vuln[1]

            vuln_types = vuln_types.union(type_of_vuln)
            vuln_lines = vuln_lines.union([node_antlr.start.line])
            vuln_columns = vuln_columns.union([node_antlr.start.column])

            # DEBUG:
            print "VULN type(s):", type_of_vuln
            print "VULN line:", node_antlr.start.line
            print "VULN column:", node_antlr.start.column
            print "VULN code:", node_antlr_text.replace("\n", "").replace("\r", "").replace("-", "").replace('"', '')
            print


class CVE_2014_4329_Sanitized_Tests(unittest.TestCase):
    ############################################################################
    def setUp(self):
        (self.dir, self.dir_realpath) = commontests.get_tests_dir_prefix()
        self.code = ""
        self.clsname = self.__class__.__name__
        self.funcname = sys._getframe().f_code.co_name

    ############################################################################
    def tearDown(self):
        self.dir = None
        self.dir_realpath = None
        self.code = ""
        self.clsname = None
        self.funcname = None

    def test_noxss_htmlescape(self):
        self.funcname = sys._getframe().f_code.co_name

        assert(self.funcname != None)
        assert(self.code != None)

        # NOTE: This is an external file with code for analysis
        host_details_lua = "./tests/3rdparty/ntopng-1.1_6932/scripts/lua/host_details.lua.htmlescape"
        test_fname = os.path.normpath(os.path.join(self.dir_realpath, host_details_lua))
        assert(test_fname != None)
        assert(os.path.exists(test_fname))
        assert(os.path.isfile(test_fname))

        luajsonres = mslparser.lua_get_json(test_fname)
        assert(luajsonres != None)

        cst_tlib = luajsonres["cst_tlib"]
        assert(cst_tlib != None)
        
        variables_all = luajsonres["variables_def"] + luajsonres["variables_local_def"]
        (variables_ordered, variables_independent, variables_dependent, variables_to_expressions_map) = taintutils.solve_variables_dependency_and_order(cst_tlib, variables_all)

        vulns_dict_all = {}

        ########################################################################
        
        for vname in variables_ordered:
            # Get the JSON object related to this variable
            v = [d for d in variables_all if d["text"] == vname][0]
            
            # Solve/taint current variable's (sub)tree
            (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, v["node"], v["text"], variables_to_expressions_map)
            
            # Merge current variable vulnerabilities dict into the all vulnerabilities dict
            vulns_dict_all.update(vulns_dict_var)
            
        ########################################################################
        
        cst_tlib_depth = cst_tlib.depth()
        all_nodes = cst_tlib.all_nodes()
        # We go just one level lower than the root node, therefore the second parameter is 0 instead of -1
        for depth in range(cst_tlib_depth, 0, -1):
            for node_tlib in all_nodes:
                node_solved = node_tlib.data.get("node_state", {"is_solved": False}).get("is_solved", False)
                if cst_tlib.level(node_tlib.identifier) == depth and not node_solved:
                    node_antlr = node_tlib.data["node_antlr"]
                    node_antlr_text = node_antlr.getText().encode('utf-8')
                    
                    # Solve/taint current unsolved node (sub)tree
                    (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, node_antlr, node_antlr_text, variables_to_expressions_map)
                    
                    # Merge current variable vulnerabilities dict into the all vulnerabilities dict
                    vulns_dict_all.update(vulns_dict_var)
                    
        ########################################################################

        vuln_types = set([])
        vuln_lines = set([])
        vuln_columns = set([])

        for vuln in vulns_dict_all.values():
            node_antlr = vuln[0]
            node_antlr_text = node_antlr.getText().encode('utf-8')
            type_of_vuln = vuln[1]

            vuln_types = vuln_types.union(type_of_vuln)
            vuln_lines = vuln_lines.union([node_antlr.start.line])
            vuln_columns = vuln_columns.union([node_antlr.start.column])

            # DEBUG:
            print "VULN type(s):", type_of_vuln
            print "VULN line:", node_antlr.start.line
            print "VULN column:", node_antlr.start.column
            print "VULN code:", node_antlr_text.replace("\n", "").replace("\r", "").replace("-", "").replace('"', '')
            print


def suite():
    suites = [CVE_2014_4329_Tests, ]
    #suites = [CVE_2014_4329_Tests, CVE_2014_4329_Sanitized_Tests, ]
    suite = unittest.TestSuite()
    for s in suites:
        suite.addTest(unittest.makeSuite(s))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')


